#include <inc/mmu.h>

.data
	// define page-aligned fsipcbuf for fsipc.c
	// ... and fdtab for file.c
	.p2align PGSHIFT
	.globl _fsipcbuf
_fsipcbuf:
	.space BY2PG
	.globl _fdtab
_fdtab:
	.space BY2PG
	.globl _envs
	.set _envs, UENVS
	.globl _pages
	.set _pages, UPAGES
	.globl _vpt
	.set _vpt, UVPT
	.globl _vpd
	.set _vpd, (UVPT+(UVPT>>12)*4)
	.globl __bss_end
__bss_end:	
	.space 4
	.set __bss_end, 0
	
.text
.globl .start
.start:
	// push empty argc, argv if none present
	cmpl $USTACKTOP, %esp
	jne args_exist
	pushl $0
	pushl $0

	// If no args, kernel set us up -- clear the bss
	// since the kernel did not.  This depends on the
	// bss not extending into a new page, since otherwise
	// we will page fault and die.
	movl $(_end), %eax
	subl $(_edata), %eax
	pushl %eax
	pushl $(_edata)
	call _bzero

	// IMPORTANT BUG FIX FOR LAB 5 -- COPY THIS NEXT LINE!
	addl $8, %esp

args_exist:
	call _libmain
1:      jmp 1b


.data
.globl __pgfault_handler
__pgfault_handler:
	.long 0

// On entry there is a trap frame on the stack:
//	[ 5 spare words ]
//	trap-time eip
//	trap-time eflags
//	trap-time esp
//	trap-time errcode
//	trap-time va		<-- %esp

.text
.globl __asm_pgfault_handler
__asm_pgfault_handler:
	// save the caller-save registers
	movl %ecx, 20(%esp)
	movl %edx, 24(%esp)
	movl %eax, 28(%esp)
	
	// call the C page fault handler
	movl __pgfault_handler, %eax
	call *%eax

	// push trap-time eip, eflags onto trap-time stack
	movl 8(%esp), %edx
	movl 12(%esp), %ecx
	movl %ecx, -4(%edx)
	movl 16(%esp), %ecx
	movl %ecx, -8(%edx)
	
	// restore the caller-save registers
	movl 20(%esp), %ecx
	movl 24(%esp), %edx
	movl 28(%esp), %eax
	
	// switch to the trap-time stack 
	movl 8(%esp), %esp

	// restore eflags and eip from the stack
	subl	$4, %esp
	popf
	jmp *-8(%esp)
