// User-level page fault handler.  We use an assembly wrapper around a C function.
// The assembly wrapper is in entry.S.

#include "lib.h"

extern void (*_pgfault_handler)(u_int, u_int);
extern void _asm_pgfault_handler(void);

//
// Set the page fault handler function.
// If there isn't one yet, _pgfault_handler will be 0.
// The first time we register a handler, we need to 
// allocate an exception stack and tell the kernel to
// call _asm_pgfault_handler on it.
//
void
set_pgfault_handler(void (*fn)(u_int va, u_int err))
{
	if (_pgfault_handler == 0) {
	  // map one page of exception stack with top at UXSTACKTOP
	  if (sys_mem_alloc(0, UXSTACKTOP-BY2PG, PTE_P|PTE_W|PTE_U)) {
	    printf("Couldn't map a page for the exception stack\n");
	    return;
	  }
	  if (sys_set_pgfault_handler(0, (u_int)_asm_pgfault_handler, UXSTACKTOP)) {
	    printf("Failed to set pgfault handler\n");
	    return;
	  }
	}

	// Save handler pointer for assembly to call.
	_pgfault_handler = fn;
}
