#include "lib.h"
#include <inc/aout.h>

static caddr_t _bss_end=0;
static int debug = 0;

static caddr_t
find_bss_end(void)
{
  struct Aout *aout = (void *)UTEXT;
  return ((caddr_t)ROUND((UTEXT+0x20+aout->a_text + aout->a_data + aout->a_bss), BY2PG));
}

caddr_t
sbrk(int incr)
{
  int r;
  caddr_t old_bss;
  
  if (_bss_end == 0)
    _bss_end = find_bss_end();

  old_bss = _bss_end;

  if (incr > 0) {
    incr = ROUND(incr, BY2PG);
  } else {
    incr = -(ROUNDDOWN(-incr, BY2PG));
  }

  while(1) {
    if (incr == 0) {
      return old_bss;
    }
    if (incr > 0) {
      if (_bss_end >= (_bss_end + incr)) {
	_bss_end = _bss_end + incr;
	return old_bss;
      }
      if ((r = sys_mem_alloc(0, (u_int)_bss_end, PTE_W|PTE_P|PTE_U)) < 0) {
	if (debug) fprintf(STDERR_FILE, "Failing to mem_alloc at 0x%x\n", _bss_end);
	return (caddr_t)r;
      }
      _bss_end += BY2PG;
      incr -= BY2PG;
    } else {
      if ((_bss_end + incr) > (caddr_t)ROUND(_bss_end-BY2PG, BY2PG)) {
	_bss_end += incr;
	return old_bss;
      }
      if ((r = sys_mem_unmap(0, ROUND(_bss_end-BY2PG, BY2PG))) < 0) {
	return (caddr_t)r;
      }
      _bss_end -= BY2PG;
      incr += BY2PG;
    }
  }
}
