#include "lib.h"
#include <inc/aout.h>
#include <stdio.h>
#include <inc/fd.h>

#define TMPPAGE		(BY2PG)
#define TMPPAGETOP	(TMPPAGE+BY2PG)

static int debug = 0;

// Is this virtual address mapped?
static u_int
va_is_mapped(u_int va)
{
	return (vpd[PDX(va)]&PTE_P) && (vpt[VPN(va)]&PTE_P);
}

static void
close_all_child_fds(u_int envid)
{
  int fdn;
  struct Fd *fd;

  for (fdn=3; fdn < MAXFD; fdn++) {
    fd = (struct Fd *)INDEX2FD(fdn);
    if (va_is_mapped((u_int)fd)) {
      // Ok this is TRULY lazy!!
      //  uses up a lot of memory that is only freed when a new file is
      // opened AND happens to be as large or larger than wasted space.
      //  could iterate through entire space to clear it.. but we are being
      //  lazy.  Better approach would be to call dev_close on the right fd,
      //  and allow dev_close an Env * argument.
      sys_mem_unmap(envid, (u_int)fd);
      sys_mem_unmap(envid, fd2data(fd));
    }
  }
}

static int
copy_library_pages(u_int envid) {
  u_int addr, addr2, pte;
  int r;

  addr = 0;
  while (addr < UTOP) {
    pte = vpd[PDX(addr)];
    if (pte & PTE_P) {
      for (addr2 = addr; addr2 < addr+PDMAP; addr2 += BY2PG) {
	pte = vpt[VPN(addr2)];
	if ((pte & PTE_P) &&
	    (pte & PTE_LIBRARY)) {
	  pte = pte & PTE_USER;
	  if ((r = sys_mem_map(0, addr2, envid, addr2, pte))) {
	    printf("Failing to map 0x%x  : %e (pte = 0x%x)\n", addr2, r, pte);
	    return r;
	  }
	}
      }
    }
    addr += PDMAP;
  }
  return 0;
}

static int
init_envp(u_int child, char **envp)
{
        int argc, r, tot, len;
	char *strings;
	u_int *args;
	u_char *strptr;

	if (envp == NULL) {
	  return -E_INVAL;
	}
	// Count the number of arguments (argc)
	// and the total amount of space needed for strings (tot)
	tot = 0;
	for (argc=0; envp[argc]; argc++)
	  tot += strlen(envp[argc])+1;

	// Make sure everything will fit in the initial stack page
	if (ROUND(tot, 4)+4*(MAX_ENV_VARS+1) > BY2PG)
	  return -E_NO_MEM;

	// Determine where to place the strings and the args array
	strings = (char*)TMPPAGE + 4*(MAX_ENV_VARS+1);
	args = (u_int*)(TMPPAGE);

	if ((r = sys_mem_alloc(0, TMPPAGE, PTE_P|PTE_U|PTE_W)) < 0)
	  return r;

	strptr = (char*)ENVBASE + 4*(MAX_ENV_VARS+1);
	for (argc=0; envp[argc]; argc++) {
	  args[argc] = (u_int)strptr;
	  strcpy(strings, envp[argc]);
	  len = strlen(strings)+1;
	  strings += len;
	  strptr += len;
	}
	args[argc] = 0;

	if ((r = sys_mem_map(0, TMPPAGE, child, ENVBASE, PTE_P|PTE_U|PTE_W)) < 0)
	  goto error;
	if ((r = sys_mem_unmap(0, TMPPAGE)) < 0)
	  goto error;

	return 0;

error:
	sys_mem_unmap(0, TMPPAGE);
	return r;
  
}
// Set up the initial stack page for the new child process with envid 'child',
// using the arguments array pointed to by 'argv',
// which is a null-terminated array of pointers to null-terminated strings.
//
// On success, returns 0 and sets *init_esp
// to the initial stack pointer with which the child should start.
// Returns < 0 on failure.
//
static int
init_stack(u_int child, char **argv, u_int *init_esp)
{
        int argc, r, tot, len;
	char *strings;
	u_int *args;
	u_char *strptr;

	// Count the number of arguments (argc)
	// and the total amount of space needed for strings (tot)
	tot = 0;
	for (argc=0; argv[argc]; argc++)
		tot += strlen(argv[argc])+1;

	// Make sure everything will fit in the initial stack page
	if (ROUND(tot, 4)+4*(argc+3) > BY2PG)
		return -E_NO_MEM;

	// Determine where to place the strings and the args array
	strings = (char*)TMPPAGETOP - tot;
	args = (u_int*)(TMPPAGETOP - ROUND(tot, 4) - 4*(argc+1));

	if ((r = sys_mem_alloc(0, TMPPAGE, PTE_P|PTE_U|PTE_W)) < 0)
		return r;

	// Replace this with your code to:
	//
	//	- copy the argument strings into the stack page at 'strings'
	//
	//	- initialize args[0..argc-1] to be pointers to these strings
	//	  that will be valid addresses for the child environment
	//	  (for whom this page will be at USTACKTOP-BY2PG!).
	//
	//	- set args[argc] to 0 to null-terminate the args array.
	strptr = (char*)USTACKTOP - tot;
	for (argc=0; argv[argc]; argc++) {
	  args[argc] = (u_int)strptr;
	  strcpy(strings, argv[argc]);
	  len = strlen(strings)+1;
	  strings += len;
	  strptr += len;
	}
	args[argc] = 0;

	//
	//	- push two more words onto the child's stack below 'args',
	//	  containing the argc and argv parameters to be passed
	//	  to the child's umain() function.
	args = (u_int *)(args - 2);
	args[0] = argc;
	args[1] = (u_int)(USTACKTOP - ROUND(tot, 4) - 4*(argc+1));
	//
	//	- set *init_esp to the initial stack pointer for the child
	//
	*init_esp = (u_int)(USTACKTOP - ROUND(tot, 4) - 4*(argc+3));	// Change this!

	if ((r = sys_mem_map(0, TMPPAGE, child, USTACKTOP-BY2PG, PTE_P|PTE_U|PTE_W)) < 0)
		goto error;
	if ((r = sys_mem_unmap(0, TMPPAGE)) < 0)
		goto error;

	return 0;

error:
	sys_mem_unmap(0, TMPPAGE);
	return r;
}

static int
open_path(char *prog)
{
  char fullpath[MAXPATHLEN];
  char *path_var;
  char *start, *end;
  int fd;

  if (strchr(prog, '/'))
    return (open(prog, O_RDONLY));

  if ((path_var = getenv("PATH")) == NULL)
    return (open(prog, O_RDONLY));

  start=path_var;
  while(*start) {
    if (*start == ':') {
      start++;
      continue;
    }
    end = (char *)strchr(start, ':');
    if (!end)
      end = start + strlen(start);
    if (end == start)
      return -E_NOT_FOUND;
    memcpy(fullpath, start, end-start);
    if (fullpath[end-start-1] != '/') {
      fullpath[end-start] = '/';
      fullpath[end-start+1] = '\0';
    } else {
      fullpath[end-start] ='\0';
    }
    strcat(fullpath, prog);
    if (debug) fprintf(STDERR_FILE, "Trying %s\n", fullpath);
    if ((fd = open(fullpath, O_RDONLY)) >= 0)
      return fd;

    start = end;
  }
  return -E_NOT_FOUND;
}

// Spawn a child process from a program image loaded from the file system.
// prog: the pathname of the program to run.
// argv: pointer to null-terminated array of pointers to strings,
// 	 which will be passed to the child as its command-line arguments.
// Returns child envid on success, < 0 on failure.
int
spawnv(char *prog, char **argv, char **envp)
{
  int fd, count, r;
  struct Aout aout_buffer;
  struct Trapframe tf;
  int envid;
  u_int esp;
  u_char *block;

  u_char *psname;

  // Insert your code, following approximately this procedure:
  //
  
  //	- Open the program file and read the a.out header (see inc/aout.h).
  if ((fd = open_path(prog)) < 0)
    return fd;

  count = read(fd, &aout_buffer, sizeof(aout_buffer));
  if (count < sizeof(aout_buffer)) {
    close(fd);
    return -E_NOT_EXEC;
  }
  //
  //	- Use sys_env_alloc() to create a new environment.
  //
  
  if (argv && argv[0])
    psname = argv[0];
  else 
    psname = prog;

  if ((envid = sys_env_alloc(psname)) < 0) {
    close(fd);
    return envid;
  }
  if (envid == 0) {
    // Really should never get here!!
    sys_env_destroy(0);
    return (envid);
  } else {
	//	- Call the init_stack() function above to set up
	//	  the initial stack page for the child environment.
	//
    if ((r = init_stack(envid, argv, &esp)) < 0) {
      // failed to set up stack...
      sys_env_destroy(envid);
      return r;
    }
    if (envp) {
      if ((r = init_envp(envid, envp)) < 0) {
	// failed to set up environment variables
	sys_env_destroy(envid);
	return r;
      }
    } else {
      if ((r = sys_mem_map(0, ENVBASE, 0, ENVBASE, PTE_P|PTE_COW|PTE_U)) < 0) {
	fprintf(STDERR_FILE, "copying environment failed\n");
	return r;
      }
      if ((r = sys_mem_map(0, ENVBASE, envid, ENVBASE, PTE_P|PTE_COW|PTE_U)) < 0) {
	fprintf(STDERR_FILE, "copying environment to child failed\n");
	return r;
      }
    }

	//	- Map the program's text segment, from file offset 0
	//	  up to (but not including) file offset 0x20+aout.a_text, starting at
	//	  virtual address UTEXT in the child (0x20 is the size of the a.out header).
	//	  Use read_map() and map the pages it returns directly
	//	  into the child so that multiple instances of the
	//	  same program will share the same copy of the program text.
	//	  Be sure to map the program text read-only in the child.
	//	  Read_map is like read but returns a pointer to the data
	//	  in *blk rather than copying the data into another buffer.
	//
	//	  The text segment will end on a page boundary.
	//	  That is, (0x20+aout.a_text) % BY2PG == 0.
	//
    for (count=0; count < 0x20 + aout_buffer.a_text; count += BY2PG) {
      if ((r = read_map(fd, count, (void *)&block)) < 0) {
	close(fd);
	sys_env_destroy(envid);
	return r;
      }
      if ((r = sys_mem_map(0, (u_int)block, envid, UTEXT+count, PTE_P|PTE_U)) < 0) {
	close(fd);
	sys_env_destroy(envid);
	return r;
      }
    }
	//	- Set up the child process's data segment.  For each page,
	//	  allocate a page in the parent temporarily at TMPPAGE,
	//	  read() the appropriate block of the file into that page,
	//	  and then insert that page mapping into the child.
	//	  Look at init_stack() for inspiration.
	//	  Be sure you understand why you can't use read_map() here.
	//
	//	  The data segment starts where the text segment left off,
	//	  so it starts on a page boundary.  The data size will always be a multiple
	//	  of the page size, so it will end on a page boundary too.
	//	  The data comes from the aout.a_data bytes in the file starting
	//	  at offset (0x20+aout.a_text).
	//
    seek(fd, 0x20+aout_buffer.a_text);
    for (count = 0; count < aout_buffer.a_data; count += BY2PG) {
      if ((r = sys_mem_alloc(0, TMPPAGE, PTE_P|PTE_U|PTE_W)) < 0) {
	close(fd);
	sys_env_destroy(envid);
	return r;
      }
      if ((r = read(fd, (void *)TMPPAGE, BY2PG)) < 0) {
	close(fd);
	sys_env_destroy(envid);
	return r;
      }
      if ((r = sys_mem_map(0, TMPPAGE, envid, UTEXT+0x20+aout_buffer.a_text+count, PTE_P|PTE_U|PTE_W)) < 0){
	close(fd);
	sys_env_destroy(envid);
	return r;
      }
      if ((r = sys_mem_unmap(0, TMPPAGE)) < 0) {
	close (fd);
	sys_env_destroy(envid);
	return r;
      }
    }
    close(fd);
      //	- Set up the child process's bss segment.
	//	  All you need to do here is sys_mem_alloc() the pages
	//	  directly into the child's address space, because
	//	  sys_mem_alloc() automatically zeroes the pages it allocates.
	//
	//	  The bss will start page aligned (since it picks up where the
	//	  data segment left off), but it's length may not be a multiple
	//	  of the page size, so it may not end on a page boundary.
	//	  Be sure to map the last page.  (It's okay to map the whole last page
	//	  even though the program will only need part of it.)
	//
	//	  The bss is not read from the binary file.  It is simply 
	//	  allocated as zeroed memory.  There are bits in the file at
	//	  offset 0x20+aout.a_text+aout.a_data, but they are not the
	//	  bss.  They are the symbol table, which is only for debuggers.
	//	  Do not use them.
	//
	//	  The exact location of the bss is a bit confusing, because
	//	  the linker lies to the loader about where it is.  
	//	  For example, in the copy of user/init that we have (yours
	//	  will depend on the size of your implementation of open and close),
	//	  i386-osclass-aout-nm claims that the bss starts at 0x8067c0
	//	  and ends at 0x807f40 (file offsets 0x67c0 to 0x7f40).
	//	  However, since this is not page aligned,
	//	  it lies to the loader, inserting some extra zeros at the end
	//	  of the data section to page-align the end, and then claims
	//	  that the data (which starts at 0x2000) is 0x5000 long, ending
	//	  at 0x7000, and that the bss is 0xf40 long, making it run from
	//	  0x7000 to 0x7f40.  This has the same effect as far as the
	//	  loading of the program.  Offsets 0x8067c0 to 0x807f40 
	//	  end up being filled with zeros, but they come from different
	//	  places -- the ones in the 0x806 page come from the binary file
	//	  as part of the data segment, but the ones in the 0x807 page
	//	  are just fresh zeroed pages not read from anywhere.
	//
	//	  If you are confused by the last paragraph, don't worry too much.
	//	  Just remember that the symbol table, should you choose to look at it,
	//	  is not likely to match what's in the a.out header.  Use the a.out
	//	  header alone.
	// 
      for (count = 0; count < aout_buffer.a_bss; count += BY2PG) {
      	if ((r = sys_mem_alloc(envid, UTEXT+0x20+aout_buffer.a_text + aout_buffer.a_data + count, PTE_U|PTE_P|PTE_W)) < 0){
	  sys_env_destroy(envid);
	  return r;
	}
      }
      bcopy(&(envs[ENVX(envid)].env_tf), &tf, (sizeof (struct Trapframe)));
      tf.tf_eip = UTEXT+0x20;
      tf.tf_esp = esp;
      if ((r = sys_set_trapframe(envid, &tf)) < 0) {
	sys_env_destroy(envid);
	return r;
      }
      if ((r = copy_library_pages(envid)) < 0) {
	printf("Failed to copy library pages\n");
	sys_env_destroy(envid);
	return r;
      }
      //  Close all randomly opened files the way exec would
      close_all_child_fds(envid);
      
      //	- Start the child process running with sys_set_env_status().
      if ((r = sys_set_env_status(envid, ENV_RUNNABLE)) < 0) {
	sys_env_destroy(envid);
	return r;
      }
      return envid;
  }
}

// Spawn, taking command-line arguments array directly on the stack.
int
spawnl(char *prog, char *args, ...)
{
	return spawn(prog, &args);
}

int
spawn(char *prog, char **argv) {
  return spawnv(prog, argv, 0);
}


